#include <libxml/parser.h>
#include <libxml/tree.h>
#include <glib-object.h>

#include "xml.h"
#include "cb.h"

#define ADJ_TYPE(value) (xmlChar *)value

void WPSStartDocumentSAXFunc(void *ctx)
{
	StateInfo *state = ctx;
	
	state->page_info = NULL;
	state->template_info = NULL;
	state->state = START_STATE;
}

void WPSEndDocumentSAXFunc(void *ctx)
{
	StateInfo *state = ctx;

	state->cur_page = NULL;
	state->cur_widget = NULL;
	state->cur_signal = NULL;
}

void WPSStartElementSAXFunc(void *ctx, const xmlChar *name, const xmlChar **attrs)
{
	StateInfo *state = ctx;
	gint i;
	
	switch (state->state) {
		case START_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("wps"))) {
				state->state = WPS_STATE;
			}
			else
				state->state = UNKNOWN_STATE; 	
			break;
			
		case WPS_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("page"))) {
				PageInfo *page;
				
				page = g_new0(PageInfo, 1);
				page->widget_info = NULL;
				page->group_widget_info = NULL;
				page->templ_use = NULL;
				page->page_prop_info = g_array_new(FALSE, TRUE, sizeof(PropInfo*));
				page->prop_len = 0;
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						page->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("description")))
						page->description = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s", attrs[i+1]);
				}
				state->cur_page = page;
				state->state = PAGE_STATE;
				g_warning("create page\n");			
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("template-def"))) {
				TemplateInfo *templ;
				
				templ = g_new0(TemplateInfo, 1);
				
				templ->group_widget_info = NULL;
				templ->widget_info = NULL;
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						templ->name = g_strdup((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("description")))
						templ->desc = g_strdup((char *)attrs[i+1]);
					else
						g_warning("strange value %s", attrs[i+1]);
				} 
				
				state->template_info = g_list_append(state->template_info, templ);
				state->state = TEMPL_STATE;
				g_warning("create template\n");
			}
			else
				state->state = UNKNOWN_STATE;	
			break;
		
		case TEMPL_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget"))) {
				WidgetInfo *widget;
				GList *last_templ;
				TemplateInfo *templ;
				
				widget = g_new0(WidgetInfo, 1);
				widget->signal_info = NULL;
				widget->style_info = NULL;
				widget->link_attributes_info = NULL;
				widget->prop_info_array = g_array_new(FALSE, TRUE, sizeof(PropInfo*));
				widget->prop_len = 0;
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						widget->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						widget->type_name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("xalign")))
						widget->xalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("yalign")))
						widget->yalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else  
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_templ = g_list_last(state->template_info);
				templ = last_templ->data;
				templ->widget_info = g_list_append(templ->widget_info, widget);
				
				state->state = TEMPL_WIDGET_STATE;
				g_warning("create widget in template\n");
			}
			else
				state->state = UNKNOWN_STATE;
			break;
		case TEMPL_WIDGET_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-property"))) {
				PropInfo *prop;
				GList *last_templ;
				GList *last_widget;
				TemplateInfo *templ;
				WidgetInfo *widget;
				
				prop = g_new0(PropInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						prop->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_templ = g_list_last(state->template_info);
				templ = last_templ->data;
				last_widget = g_list_last(templ->widget_info);
				widget = last_widget->data;
				
				g_array_append_val(widget->prop_info_array, prop);
				state->state = TEMPL_WIDGET_PROPERTY_STATE;
				
				g_warning("set widget properties template\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("signal"))) {
				SignalInfo *signal;
				GList *last_templ;
				GList *last_widget;
				WidgetInfo *widget;
				TemplateInfo *templ;
				
				signal = g_new0(SignalInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("handler")))
						signal->handler = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_templ = g_list_last(state->template_info);
				templ = last_templ->data;
				last_widget = g_list_last(templ->widget_info);
				widget = last_widget->data;
				widget->signal_info = g_list_append(widget->signal_info, signal);
				
				state->state = TEMPL_WIDGET_SIGNAL_STATE;
				g_warning("set widget properties template\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("style"))) {
				StyleInfo *style;
				GList *last_templ;
				GList *last_widget;
				WidgetInfo *widget;
				TemplateInfo *templ;
				style = g_new0(StyleInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("font-size")))
						style->font_size = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-color")))
						style->field_color = g_strdup((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-font-color")))
						style->field_font_color = g_strdup((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_templ = g_list_last(state->template_info);
				templ = last_templ->data;
				last_widget = g_list_last(templ->widget_info);
				widget = last_widget->data;
				widget->style_info = style;
				
				state->state = TEMPL_WIDGET_STYLE_STATE;	
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("link-attribute"))) {
				LinkAttributeInfo *link_attribute;
				GList *last_templ;
				GList *last_widget;
				WidgetInfo *widget;
				TemplateInfo *templ;
				
				link_attribute = g_new0(LinkAttributeInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("tag")))
						link_attribute->tag = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						link_attribute->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						link_attribute->type = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("direction")))
						link_attribute->direction = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i+1]);
				}
				
				last_templ = g_list_last(state->template_info);
				templ = last_templ->data;
				last_widget = g_list_last(templ->widget_info);
				widget = last_widget->data;
				
				widget->link_attributes_info = g_list_append(widget->link_attributes_info, link_attribute);
				state->state = TEMPL_WIDGET_LINK_ATTRIBUTE_STATE;
			}
			else
				state->state = UNKNOWN_STATE;
			break;
			
		case PAGE_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget"))) {
				WidgetInfo *widget;
				
				widget = g_new0(WidgetInfo, 1);
				widget->signal_info = NULL;
				widget->style_info = NULL;
				widget->link_attributes_info = NULL;
				widget->prop_info_array = g_array_new(FALSE, TRUE, sizeof(PropInfo*));
				widget->prop_len = 0;
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						widget->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						widget->type_name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("xalign")))
						widget->xalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("yalign")))
						widget->yalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else  
						g_warning("strange value %s\n", attrs[i]);
				}
				state->cur_widget = widget;
				state->state = WIDGET_STATE;
				g_warning("create widget\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("widget-group"))) {
				GroupWidgetInfo *group;
				
				group = g_new0(GroupWidgetInfo, 1);
				group->prop_info_array = g_array_new(FALSE, TRUE, sizeof(PropInfo*));
				group->prop_len = 0;
								
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						group->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]); 
				}
				
				state->cur_page->group_widget_info = g_list_append(state->cur_page->group_widget_info, group);
				state->state = GROUP_STATE;
				g_warning("set new group\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("page-property"))) {
				PropInfo *prop;
				prop = g_new0(PropInfo, 1);
								
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						prop->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				g_array_append_val(state->cur_page->page_prop_info, prop);
				state->state = PAGE_PROPERTY_STATE;
				g_warning("set page properties\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("template-use"))) {
				TemplateInUse *templ_use;
				PageInfo *page;
				
				templ_use = g_new0(TemplateInUse, 1);
								
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						templ_use->name = g_strdup((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("xalign")))
						templ_use->xalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("yalign")))
						templ_use->yalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				page = state->cur_page;
				page->templ_use = g_list_append(page->templ_use, templ_use);
				
				state->state = PAGE_TEMPL;
			}
			else
				state->state = UNKNOWN_STATE;
			break;
		case WIDGET_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-property"))) {
				PropInfo *prop;
				prop = g_new0(PropInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						prop->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				g_array_append_val(state->cur_widget->prop_info_array, prop);
				state->state = WIDGET_PROPERTY_STATE;
				
				g_warning("set widget properties\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("signal"))) {
				SignalInfo *signal;
				signal = g_new0(SignalInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("handler")))
						signal->handler = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				state->cur_signal = signal;
				state->state = SIGNAL_STATE;
				g_warning("set widget properties\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("style"))) {
				StyleInfo *style;
				style = g_new0(StyleInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("font-size")))
						style->font_size = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-color")))
						style->field_color = g_strdup((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-font-color")))
						style->field_font_color = g_strdup((char *)attrs[i+1]); 
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				state->cur_widget->style_info = style;
				state->state = WIDGET_STYLE_STATE;	
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("link-attribute"))) {
				LinkAttributeInfo *link_attribute;
				
				link_attribute = g_new0(LinkAttributeInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("tag")))
						link_attribute->tag = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						link_attribute->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						link_attribute->type = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("direction")))
						link_attribute->direction = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i+1]);
				}
				
				state->cur_widget->link_attributes_info = g_list_append(state->cur_widget->link_attributes_info, link_attribute);
				state->state = WIDGET_LINK_ATTRIBUTE_STATE;
			}
			else
				state->state = UNKNOWN_STATE;
			break;
		case GROUP_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget"))) {
				WidgetInfo *widget;
				GroupWidgetInfo *group;
				GList *last_group;
				
				widget = g_new0(WidgetInfo, 1);
				widget->signal_info = NULL;
				widget->prop_info_array = g_array_new(FALSE, TRUE, sizeof(PropInfo*));
				widget->prop_len = 0;
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						widget->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						widget->type_name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("xalign")))
						widget->xalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("yalign")))
						widget->yalign = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else  
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				group->widget_info = g_list_append(group->widget_info, widget);
				
				g_warning("%s\n", widget->name->str);
				state->state = GROUP_WIDGET_STATE;
				g_warning("create widget in group\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("group-property"))) {
				PropInfo *prop;
				GList *last_group;
				GroupWidgetInfo *group;
				
				prop = g_new0(PropInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						prop->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				g_array_append_val(group->prop_info_array, prop);
				state->state = GROUP_PROPERTY_STATE;
				
				g_warning("set group properties\n");
				
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("signal"))) {
				SignalInfo *signal;
				GList *last_group = NULL;
				GroupWidgetInfo *group;
				
				signal = g_new0(SignalInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("handler")))
						signal->handler = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				group->signal_info = g_list_append(group->signal_info, signal);
				state->state = GROUP_SIGNAL_STATE;
				g_warning("set widget SIGNAL\n");
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("link-attribute"))) {
				LinkAttributeInfo *link_attribute;
				GList *last_group = NULL;
				GroupWidgetInfo *group;
				
				link_attribute = g_new0(LinkAttributeInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("tag")))
						link_attribute->tag = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						link_attribute->name = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("type")))
						link_attribute->type = g_string_new((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("direction")))
						link_attribute->direction = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i+1]);
				}
				
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				group->link_attributes_info = g_list_append(group->link_attributes_info, link_attribute);
				state->state = GROUP_LINK_ATTRIBUTE_STATE;
				g_warning("set widget LINK-ATTR\n");
			}
			else
				state->state = UNKNOWN_STATE;
			break;
		case GROUP_PROPERTY_STATE:
			state->prev_state = state->state;
			state->state = UNKNOWN_STATE;
			state->unknown_depth++;
			break;
		case GROUP_WIDGET_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-property"))) {
				PropInfo *prop;
				GroupWidgetInfo *group;
				WidgetInfo *widget;
				GList *last_group, *last_widget;
				
				prop = g_new0(PropInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("name")))
						prop->name = g_string_new((char *)attrs[i+1]);
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				last_widget = g_list_last(group->widget_info);
				widget = last_widget->data;
				g_array_append_val(widget->prop_info_array, prop);
				state->state = GROUP_WIDGET_PROPERTY_STATE;
				g_warning("set page widget properties %s\n", prop->name->str);
			}
			else if (!xmlStrcmp(name, ADJ_TYPE("style"))) {
				StyleInfo *style;
				GList *last_group, *last_widget;
				GroupWidgetInfo *group;
				WidgetInfo *widget;
				
				style = g_new0(StyleInfo, 1);
				
				for (i = 0; attrs && attrs[i] != 0; i += 2) {
					if (!xmlStrcmp(attrs[i], ADJ_TYPE("font-size")))
						style->font_size = g_ascii_strtoll((char *)attrs[i+1], NULL, 10);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-color")))
						style->field_color = g_strdup((char *)attrs[i+1]);
					else if (!xmlStrcmp(attrs[i], ADJ_TYPE("field-font-color")))
						style->field_font_color = g_strdup((char *)attrs[i+1]); 
					else
						g_warning("strange value %s\n", attrs[i]);
				}
				last_group = g_list_last(state->cur_page->group_widget_info);
				group = last_group->data;
				last_widget = g_list_last(group->widget_info);
				widget = last_widget->data;
				widget->style_info = style;
				state->state = GROUP_WIDGET_STYLE_STATE;	
			}
			else
				state->state = UNKNOWN_STATE;
			break;
		case PAGE_PROPERTY_STATE:
			/*state->prev_state = state->state;
			state->state = UNKNOWN_STATE;
			state->unknown_depth++;
			break;*/	
		case GROUP_WIDGET_PROPERTY_STATE:
			/*state->prev_state = state->state;
			state->state = UNKNOWN_STATE;
			state->unknown_depth++;
			break;*/	
		case WIDGET_PROPERTY_STATE:
			/*state->prev_state = state->state;
			state->state = UNKNOWN_STATE;
			state->unknown_depth++;
			break;*/
		case SIGNAL_STATE:
		case WIDGET_STYLE_STATE:
		case GROUP_WIDGET_STYLE_STATE:
		case WIDGET_LINK_ATTRIBUTE_STATE:
		case TEMPL_WIDGET_PROPERTY_STATE:
		case TEMPL_WIDGET_SIGNAL_STATE:
		case TEMPL_WIDGET_STYLE_STATE:
		case TEMPL_WIDGET_LINK_ATTRIBUTE_STATE:
		case PAGE_TEMPL:
		case GROUP_SIGNAL_STATE:
		case GROUP_LINK_ATTRIBUTE_STATE:
			state->prev_state = state->state;
			state->state = UNKNOWN_STATE;
			state->unknown_depth++;
			break;
		
		case UNKNOWN_STATE:
			state->unknown_depth++;
			break;
		
		default:
			break;
	}
	
}

void WPSEndElementSAXFunc(void *ctx, const xmlChar *name)
{
	StateInfo *state = ctx;	
	
	switch (state->state) {
		case START_STATE:
			break;
			
		case WPS_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("wps")) != 0)
				g_warning("waiting for </wps>");
			
			state->state = FINISH_STATE;
			break;
			
		case PAGE_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("page")) != 0)
				g_warning("waiting for </page>");
			
			state->page_info = g_list_append(state->page_info, state->cur_page);
			state->state = WPS_STATE;
			break;
			
		case WIDGET_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("widget")) != 0)
				g_warning("waiting for </widget>");
			
			state->cur_page->widget_info = g_list_append(state->cur_page->widget_info, state->cur_widget);
			state->state = PAGE_STATE;
			break;
			
		case PAGE_PROPERTY_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("page-property")) != 0)
				g_warning("waiting for </page-property>");
			state->state = PAGE_STATE;
			break;
		case PAGE_TEMPL:
			if (xmlStrcmp(name, ADJ_TYPE("template-use")) != 0)
				g_warning("waiting for </template-use>");
			state->state = PAGE_STATE;
			break;
			
		case TEMPL_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("template-def")) != 0)
				g_warning("waiting for </template-def>");

			state->state = WPS_STATE;
			break;
		case TEMPL_WIDGET_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("widget")) != 0)
				g_warning("waiting for </widget>");

			state->state = TEMPL_STATE;
			break;				
		case TEMPL_WIDGET_PROPERTY_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("widget-property")) != 0)
				g_warning("waiting for </page-property>");
			state->state = TEMPL_WIDGET_STATE;
			break;
		case TEMPL_WIDGET_SIGNAL_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("signal")) != 0)
				g_warning("waiting for </page-property>");
			state->state = TEMPL_WIDGET_STATE;
			break;
		case TEMPL_WIDGET_STYLE_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("style")) != 0)
				g_warning("waiting for </page-property>");
			state->state = TEMPL_WIDGET_STATE;
			break;
		case TEMPL_WIDGET_LINK_ATTRIBUTE_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("widget-property")) != 0)
				g_warning("waiting for </widget-property>");
			state->state = TEMPL_WIDGET_STATE;
			break;
		case WIDGET_PROPERTY_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-property")) == 0)
				g_warning("waiting for </widget-property>");

			state->state = WIDGET_STATE;
			break;
		case WIDGET_LINK_ATTRIBUTE_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("link-attribute")) == 0)
				g_warning("waiting for </widget-property>");
				
			state->state = WIDGET_STATE;
			break;
		case SIGNAL_STATE:
			if (xmlStrcmp(name, ADJ_TYPE("signal")) != 0)
				g_warning("waiting for </signal>");
			
			add_key_to_hash(state->cur_signal->handler->str);
			state->cur_widget->signal_info = g_list_append(state->cur_widget->signal_info, state->cur_signal);
			state->state = WIDGET_STATE;
			break;
		case WIDGET_STYLE_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("style")) == 0)
				g_warning("waiting for </style>");

			state->state = WIDGET_STATE;
			break;
		case GROUP_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-group")) == 0)
				g_warning("waiting for </widget-group>");

			state->state = PAGE_STATE;
			break;
		case GROUP_WIDGET_STYLE_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("style")) == 0)
				g_warning("waiting for </style>");

			state->state = GROUP_WIDGET_STATE;
			break;
		case GROUP_SIGNAL_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("signal")) == 0)
				g_warning("waiting for </signal>");

			state->state = GROUP_STATE;
			break;
		case GROUP_LINK_ATTRIBUTE_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("link-attribute")) == 0)
				g_warning("waiting for </link-attribute>");

			state->state = GROUP_STATE;
			break;
		case GROUP_WIDGET_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget")) == 0)
				g_warning("waiting for </widget>");

			state->state = GROUP_STATE;
			break;
		case GROUP_PROPERTY_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("group-property")) == 0)
				g_warning("waiting for </group-property>");

			state->state = GROUP_STATE;
			break;
		case GROUP_WIDGET_PROPERTY_STATE:
			if (!xmlStrcmp(name, ADJ_TYPE("widget-property")) == 0)
				g_warning("waiting for </widget-property>");

			state->state = GROUP_WIDGET_STATE;
			break;
		case FINISH_STATE:
			break;
			
		case UNKNOWN_STATE:
			state->unknown_depth--;
		
			if (state->unknown_depth == 0)
	    			state->state = state->prev_state;
			break;
						
		default:
			break;
	}
}

void WPSCharacters(void *ctx, const xmlChar *ch, int len)
{
	StateInfo *state = ctx;
	PropInfo *prop;
	GroupWidgetInfo *group;
	WidgetInfo *widget;
	TemplateInfo *templ;
	GList *last_group, *last_widget, *last_templ;
	gint i;
	gchar *tmp;
	
	switch (state->state) {
		case WIDGET_PROPERTY_STATE:
			tmp = g_strdup((gchar *)ch);
			prop = g_array_index(state->cur_widget->prop_info_array,
						PropInfo*,
						state->cur_widget->prop_len);

			for (i = 0; i < len; i++) {
				if (tmp[i] == '[')
					tmp[i] = '<';
				else if (tmp[i] == ']')
					tmp[i] = '>';
			}
			prop->value = g_string_new_len((char *)tmp, len);
			state->cur_widget->prop_len++;
			break;
		case PAGE_PROPERTY_STATE:
			prop = g_array_index(state->cur_page->page_prop_info,
						PropInfo*,
						state->cur_page->prop_len);
			prop->value = g_string_new_len((char *)ch, len);
			state->cur_page->prop_len++;
			break;
		case GROUP_WIDGET_PROPERTY_STATE:
			tmp = g_strdup((gchar *)ch);
			last_group = g_list_last(state->cur_page->group_widget_info);
			group = last_group->data;
			last_widget = g_list_last(group->widget_info);
			widget = last_widget->data;
			prop = g_array_index(widget->prop_info_array,
						PropInfo*,
						widget->prop_len);
						
			for (i = 0; i < len; i++) {
				if (tmp[i] == '[')
					tmp[i] = '<';
				else if (tmp[i] == ']')
					tmp[i] = '>';
			}
			prop->value = g_string_new_len((char *)tmp, len);
			widget->prop_len++;
			break;
		case TEMPL_WIDGET_PROPERTY_STATE:
			tmp = g_strdup((gchar *)ch);
			last_templ = g_list_last(state->template_info);
			templ = last_templ->data;
			last_widget = g_list_last(templ->widget_info);
			widget = last_widget->data;
			prop = g_array_index(widget->prop_info_array,
						PropInfo*,
						widget->prop_len);
						
			for (i = 0; i < len; i++) {
				if (tmp[i] == '[')
					tmp[i] = '<';
				else if (tmp[i] == ']')
					tmp[i] = '>';
			}
			prop->value = g_string_new_len((char *)tmp, len);
			widget->prop_len++;
			break;
		case GROUP_PROPERTY_STATE:
			last_group = g_list_last(state->cur_page->group_widget_info);
			group = last_group->data;
			prop = g_array_index(group->prop_info_array,
						PropInfo*,
						group->prop_len);
			prop->value = g_string_new_len((char *)ch, len);
			group->prop_len++;
			break;
		default:
			break;
	}
}

xmlSAXHandler sax_parser = {
	.startDocument = WPSStartDocumentSAXFunc,
	.endDocument = WPSEndDocumentSAXFunc,
	.startElement = WPSStartElementSAXFunc,
	.endElement = WPSEndElementSAXFunc,
	.characters = WPSCharacters
};

ResultInfo *parse_xml_file(gchar *file)
{
	StateInfo state;
	ResultInfo *result;
	
	result = g_new0(ResultInfo, 1);
	
	if (xmlSAXUserParseFile(&sax_parser, &state, file) != 0) {
		g_warning("Error during parsing file\n");
		exit(0);
	}
	
	if (state.state != FINISH_STATE) {
		g_warning("Fail state after parsing file %d\n", state.state);
		exit(0);
	}
	
	result->page_info = state.page_info;
	result->template_info = state.template_info;
	
	return result;
}

